package org.artifactory.queue.error;

import lombok.AllArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.artifactory.queue.QueueConfiguration;
import org.artifactory.queue.model.PersistentQueueRunningContext;
import org.artifactory.queue.model.PersistentQueueRunningResult;
import org.artifactory.queue.register.RepoAndCursor;
import org.artifactory.queue.services.PersistentQueue;
import org.artifactory.storage.StorageException;
import org.artifactory.storage.db.event.entity.NodeEventCursorType;
import org.artifactory.storage.db.event.error.dao.EventAndErrorInfoDao;
import org.artifactory.storage.db.event.error.factory.NodeEventsErrorInfoFactory;
import org.artifactory.storage.db.event.error.model.EventAndErrorInfo;
import org.artifactory.storage.db.event.error.model.NodeEventsErrorInfo;
import org.artifactory.storage.db.event.error.service.NodeEventsErrorsStorageService;
import org.artifactory.storage.event.EventInfo;
import org.jetbrains.annotations.NotNull;
import org.jfrog.common.ClockUtils;
import org.springframework.stereotype.Service;

import javax.annotation.Nonnull;
import java.sql.SQLException;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.stream.Collectors;

/**
 * @author avishayh
 */
@AllArgsConstructor
@Service
@Slf4j
public class PersistentQueueErrorServiceImpl implements PersistentQueueErrorService {

    private final NodeEventsErrorsStorageService errorsStorageService;
    private final EventAndErrorInfoDao errorInfoDao;

    @Override
    public int getErrorCount(PersistentQueue queue) {
        QueueConfiguration queueConfiguration = queue.getQueueConfiguration();
        if (!queueConfiguration.isEnableErrorSupport()) {
            return 0;
        }
        String operationKey = queue.getName();
        String eventCategory = queueConfiguration.getCursorType().getName();
        try {
            return errorInfoDao.getErrorCount(eventCategory, operationKey);
        } catch (SQLException exception) {
            throw new StorageException("Failed to retrieve error count", exception);
        }
    }

    @Override
    @Nonnull
    public List<EventAndErrorInfo> getErrors(PersistentQueue queue) {
        QueueConfiguration queueConfiguration = queue.getQueueConfiguration();
        if (!queueConfiguration.isEnableErrorSupport()) {
            //nothing to return if disabled
            return Collections.emptyList();
        }
        String operationKey = queue.getName();
        String eventCategory = queueConfiguration.getCursorType().getName();
        try {
            return errorInfoDao.loadErrorEvents(eventCategory, operationKey, -1);
        } catch (SQLException e) {
            throw new StorageException("Failed to retrieve errors", e);
        }
    }

    @Override
    public void processErrors(PersistentQueue queue, PersistentQueueRunningContext context) {
        if (!queue.getQueueConfiguration().isEnableErrorSupport()) {
            return;
        }
        Map<Long, EventAndErrorInfo> existingErrors = context.getExistingErrors();
        Set<Long> failed = context.getFailedResults().stream()
                .map(PersistentQueueRunningResult::getFailedMarker)
                .filter(marker -> marker > -1)
                .collect(Collectors.toSet());

        long now = ClockUtils.epochMillis();

        //merge to original failed markers and the new failed markers
        Set<Long> markers = new HashSet<>();
        markers.addAll(existingErrors.keySet());
        markers.addAll(failed);
        for (long marker : markers) {
            boolean existedBefore = existingErrors.containsKey(marker);
            if (existedBefore) {
                updateOrDeleteError(marker, queue, context);
            } else {
                createNewError(marker, queue, context, now);
            }
        }
    }

    @Override
    public void deleteErrors(PersistentQueue queue) {
        QueueConfiguration queueConfiguration = queue.getQueueConfiguration();
        if (!queueConfiguration.isEnableErrorSupport()) {
            return;
        }
        String operationKey = queue.getName();
        String eventCategory = queueConfiguration.getCursorType().getName();
        try {
            errorsStorageService.deleteAll(eventCategory, operationKey);
        } catch (Exception exception) {
            log.error("Failed cleanup errors for queue: {}, error: {}", queue.getName(), exception);
        }
    }

    @Override
    public void deleteErrors(@Nonnull Collection<Long> errors) {
        errors.forEach(error -> errorsStorageService.delete(error));
    }

    @NotNull
    @Override
    public Set<RepoAndCursor> getErroneousQueues() {
        return errorsStorageService.getErroneousRepos().stream()
                .map(e -> new RepoAndCursor(e.getRepoKey(), NodeEventCursorType.valueOf(e.getEventCategory())))
                .collect(Collectors.toSet());
    }

    private void updateOrDeleteError(long failedMarker, PersistentQueue queue, PersistentQueueRunningContext context) {
        QueueConfiguration queueConfiguration = queue.getQueueConfiguration();
        EventAndErrorInfo existingError = context.getExistingErrors().get(failedMarker);

        PersistentQueueRunningResult failedResult = context.getFailedResults().stream()
                .filter(r -> r.getFailedMarker() == failedMarker)
                .findFirst().orElse(null);

        boolean errorResolved = failedResult == null;
        boolean errorExhausted = existingError.getErrorCount() + 1 >= queueConfiguration.getMaxErrorRetires();
        if (errorExhausted) {
            // Add the failedEventId to the context, so we could consider it when calculating
            // the TriggerResult
            context.getExhaustedErrors().add(failedMarker);
            log.warn("Too much retries ({}) for event {}. event is exhausted",
                    existingError.getErrorCount() + 1,
                    existingError.getEventInfo());
        }
        if (errorResolved || errorExhausted) {
            errorsStorageService.delete(existingError.getErrorId());
        } else {
            EventAndErrorInfo updatedError = NodeEventsErrorInfoFactory.updateExisting(existingError);
            errorsStorageService.updateById(toErrorInfo(updatedError, queue, failedResult.getErrorMessage()),
                    updatedError.getErrorId());
        }
    }

    private void createNewError(long failedMarker, PersistentQueue queue, PersistentQueueRunningContext context,
            long now) {

        //failed result must exist now
        PersistentQueueRunningResult failedResult = context.getFailedResults().stream()
                .filter(r -> r.getFailedMarker() == failedMarker)
                .findFirst().orElseThrow(() -> new IllegalStateException("result must exist"));

        //find the event
        EventInfo eventInfo =
                context.getEvents().stream().filter(e -> e.getEventId() == failedMarker)
                        .findFirst()
                        .orElseThrow(() -> new IllegalStateException("marker must exist"));
        EventAndErrorInfo error = NodeEventsErrorInfoFactory.createNew(eventInfo, now);
        errorsStorageService.add(toErrorInfo(error, queue, failedResult.getErrorMessage()));
    }

    private NodeEventsErrorInfo toErrorInfo(EventAndErrorInfo error, PersistentQueue queue, String errorMessage) {
        QueueConfiguration queueConfiguration = queue.getQueueConfiguration();
        String operationKey = queue.getName();
        String eventCategory = queueConfiguration.getCursorType().getName();
        return NodeEventsErrorInfo.builder()
                .eventType(error.getEventInfo().getType())
                .path(error.getEventInfo().getRepoPath())
                .eventId(error.getEventInfo().getEventId())
                .eventTime(error.getEventInfo().getTimestamp())
                .firstErrorTime(error.getFirstErrorTime())
                .errorCount(error.getErrorCount())
                .eventCategory(eventCategory)
                .operationKey(operationKey)
                .errorMessage(errorMessage)
                .build();
    }
}
