package org.artifactory.security.access;

import com.google.common.collect.ImmutableList;
import org.artifactory.addon.AddonsManager;
import org.artifactory.addon.LdapGroupAddon;
import org.artifactory.api.config.CentralConfigService;
import org.artifactory.api.security.UserGroupService;
import org.artifactory.api.security.ldap.LdapUser;
import org.artifactory.api.security.ldap.LdapUserAttributes;
import org.artifactory.descriptor.config.MutableCentralConfigDescriptor;
import org.artifactory.descriptor.security.SecurityDescriptor;
import org.artifactory.descriptor.security.ldap.LdapSetting;
import org.artifactory.model.xstream.security.GroupImpl;
import org.artifactory.model.xstream.security.UserImpl;
import org.artifactory.security.MutableUserInfo;
import org.artifactory.security.ldap.LdapServiceImpl;
import org.artifactory.test.ArtifactoryHomeBoundTest;
import org.jetbrains.annotations.NotNull;
import org.jfrog.access.client.token.TokenVerifyResult;
import org.jfrog.access.common.ServiceId;
import org.jfrog.access.model.Realm;
import org.jfrog.access.token.JwtAccessToken;
import org.mockito.Mock;
import org.mockito.MockitoAnnotations;
import org.springframework.ldap.core.DirContextOperations;
import org.springframework.ldap.core.LdapTemplate;
import org.springframework.security.authentication.InternalAuthenticationServiceException;
import org.springframework.security.core.Authentication;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.DataProvider;
import org.testng.annotations.Test;

import javax.naming.directory.Attributes;
import java.util.List;

import static org.mockito.ArgumentMatchers.any;
import static org.mockito.ArgumentMatchers.eq;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;
import static org.testng.Assert.assertTrue;
import static org.testng.AssertJUnit.assertEquals;
import static org.testng.AssertJUnit.assertFalse;

/**
 * @author Shay Bagants
 */
@Test
public class AccessTokenAuthenticationProviderTest extends ArtifactoryHomeBoundTest {

    private AccessTokenAuthenticationProvider tokenAuthenticationProvider;

    @Mock
    private AccessService accessService;
    @Mock
    private UserGroupService userGroupService;
    @Mock
    private CentralConfigService centralConfigService;
    @Mock
    private LdapServiceImpl ldapService;
    @Mock
    private AddonsManager addonsManager;

    @BeforeMethod
    public void setup() {
        MockitoAnnotations.initMocks(this);

        tokenAuthenticationProvider = new AccessTokenAuthenticationProvider();
        tokenAuthenticationProvider.setAccessService(accessService);
        tokenAuthenticationProvider.setUserGroupService(userGroupService);
        tokenAuthenticationProvider.setCentralConfig(centralConfigService);
        tokenAuthenticationProvider.setLdapService(ldapService);
        tokenAuthenticationProvider.setAddonsManager(addonsManager);

        ServiceId serviceId = new ServiceId("jfrt", "xyz");
        when(accessService.getArtifactoryServiceId()).thenReturn(serviceId);
    }

    @Test(dataProvider = "provideScopes")
    public void testSetAdminPrivileges(List<String> scopes, boolean expectedAdmin) {
        UserImpl user = new UserImpl();
        JwtAccessToken token = mock(JwtAccessToken.class);
        when(token.getScope()).thenReturn(scopes);
        tokenAuthenticationProvider.setAdminPrivileges(user, new UserImpl(), token);
        assertEquals(user.isAdmin(), expectedAdmin);
    }

    public void testSetAdminPrivilegesFromOriginalUser() {
        UserImpl user = new UserImpl();
        UserImpl userFromStorage = new UserImpl();
        userFromStorage.setAdmin(true);
        JwtAccessToken token = mock(JwtAccessToken.class);
        when(token.getScope()).thenReturn(ImmutableList.of("applied-permissions/user"));
        tokenAuthenticationProvider.setAdminPrivileges(user, userFromStorage, token);
        assertTrue(user.isAdmin());

        user = new UserImpl();
        when(token.getScope()).thenReturn(ImmutableList.of("applied-permissions/groups:group1"));
        tokenAuthenticationProvider.setAdminPrivileges(user, userFromStorage, token);
        // not an identity token, no admin rights
        assertFalse(user.isAdmin());

        user = new UserImpl();
        when(token.getScope())
                .thenReturn(ImmutableList.of("applied-permissions/user", "applied-permissions/groups:group1"));
        tokenAuthenticationProvider.setAdminPrivileges(user, userFromStorage, token);
        // both identity and non-identity scopes are combined
        assertTrue(user.isAdmin());
    }

    @Test(enabled = false, description = "Flaky. Need to investigate")
    public void testGroupsCombinedFromIdentityAndNonIdentityScopes() {
        UserImpl user = new UserImpl();
        user.setUsername("joe");
        UserImpl userFromStorage = new UserImpl();
        userFromStorage.setUsername("joe");
        JwtAccessToken token = mock(JwtAccessToken.class);

        when(token.getScope())
                .thenReturn(ImmutableList.of("applied-permissions/user", "applied-permissions/groups:group1"));
        when(userGroupService.getAllGroups())
                .thenReturn(ImmutableList.of(new GroupImpl("group1"), new GroupImpl("notARealGroup")));

        MutableUserInfo mutatedUser = tokenAuthenticationProvider.populateGroups(user, userFromStorage, token);

        assertTrue(mutatedUser.getGroups().stream().anyMatch(group -> "group1".equals(group.getGroupName())));
        assertTrue(mutatedUser.getGroups().stream().noneMatch(group -> "notARealGroup".equals(group.getGroupName())));
    }

    public void testNonLdapAuthenticate() {
        final String userName = "ldapadmin";
        JwtAccessToken token = mockToken(userName);
        final TokenVerifyResult verifyResult = mock(TokenVerifyResult.class);
        when(verifyResult.isSuccessful()).thenReturn(true);
        when(accessService.verifyAndGetResult(token)).thenReturn(verifyResult);
        when(accessService.extractSubjectUsername(token)).thenReturn(userName);
        final Authentication authenticate =
                tokenAuthenticationProvider.authenticate(new AccessTokenAuthentication(token, userName, List.of()));
        assertTrue(authenticate.isAuthenticated());
    }

    @Test(expectedExceptions = InternalAuthenticationServiceException.class,
            expectedExceptionsMessageRegExp = ".*'ldapadmin': user is locked, disabled or does not exist in LDAP")
    public void testLdapAuthenticationFailed() {
        final String userName = "ldapadmin";
        UserImpl userFromStorage = new UserImpl();
        userFromStorage.setRealm(Realm.LDAP.getName());
        userFromStorage.setAdmin(true);
        JwtAccessToken token = mockToken(userName);
        final LdapSetting ldapSetting = new LdapSetting();
        ldapSetting.setEnabled(false);
        mockLdapSettings(ldapSetting);
        final TokenVerifyResult verifyResult = mock(TokenVerifyResult.class);
        when(verifyResult.isSuccessful()).thenReturn(true);
        when(accessService.verifyAndGetResult(token)).thenReturn(verifyResult);
        when(accessService.extractSubjectUsername(token)).thenReturn(userName);
        when(userGroupService.findUser(userName, false)).thenReturn(userFromStorage);
        when(userGroupService.findUser(userName)).thenReturn(userFromStorage);

        tokenAuthenticationProvider.authenticate(new AccessTokenAuthentication(token, userName, List.of()));
    }

    public void testLdapAuthenticationSucceeded() {
        final String userName = "ldapadmin";
        UserImpl userFromStorage = new UserImpl();
        userFromStorage.setRealm(Realm.LDAP.getName());
        userFromStorage.setAdmin(true);
        JwtAccessToken token = mockToken(userName);
        final LdapSetting ldapSetting = new LdapSetting();
        ldapSetting.setEnabled(true);
        mockLdapSettings(ldapSetting);
        final TokenVerifyResult verifyResult = mock(TokenVerifyResult.class);
        when(verifyResult.isSuccessful()).thenReturn(true);
        when(accessService.verifyAndGetResult(token)).thenReturn(verifyResult);
        when(accessService.extractSubjectUsername(token)).thenReturn(userName);
        when(userGroupService.findUser(userName, false)).thenReturn(userFromStorage);
        when(userGroupService.findUser(userName)).thenReturn(userFromStorage);
        when(userGroupService.findOrCreateExternalAuthUser(userName, false, false)).thenReturn(userFromStorage);

        final LdapUserAttributes ldapUser = mockLdapUser();
        when(ldapService.getUserWithAttributesFromDomain(any(), eq(userName))).thenReturn(ldapUser);
        final LdapTemplate ldapTemplate = mock(LdapTemplate.class);
        when(ldapService.createLdapTemplate(any())).thenReturn(ldapTemplate);
        final DirContextOperations contextOperations = mock(DirContextOperations.class);
        when(ldapService.searchUserInLdap(ldapTemplate, userName, ldapSetting)).thenReturn(contextOperations);
        final LdapGroupAddon groupAddon = mock(LdapGroupAddon.class);
        when(addonsManager.addonByType(LdapGroupAddon.class)).thenReturn(groupAddon);
        final Authentication authenticate =
                tokenAuthenticationProvider.authenticate(new AccessTokenAuthentication(token, userName, List.of()));
        assertTrue(authenticate.isAuthenticated());
    }

    @NotNull
    private LdapUserAttributes mockLdapUser() {
        final LdapUserAttributes ldapUser = mock(LdapUserAttributes.class);
        final Attributes attributes = mock(Attributes.class);
        final LdapUser ldapUserEntity = mock(LdapUser.class);
        when(ldapUser.getAttributes()).thenReturn(attributes);
        when(ldapUser.getLdapUser()).thenReturn(ldapUserEntity);
        return ldapUser;
    }

    private void mockLdapSettings(LdapSetting ldapSetting) {
        final MutableCentralConfigDescriptor mutableCentralConfig = mock(MutableCentralConfigDescriptor.class);
        when(centralConfigService.getMutableDescriptor()).thenReturn(mutableCentralConfig);
        final SecurityDescriptor securityDescriptor = mock(SecurityDescriptor.class);
        when(mutableCentralConfig.getSecurity()).thenReturn(securityDescriptor);
        when(securityDescriptor.getLdapSettings()).thenReturn(List.of(ldapSetting));
    }

    @NotNull
    private JwtAccessToken mockToken(String userName) {
        JwtAccessToken token = mock(JwtAccessToken.class);
        when(token.getIssuer()).thenReturn(userName);
        when(token.getSubject()).thenReturn(userName);
        return token;
    }

    @DataProvider
    public static Object[][] provideScopes() {
        return new Object[][] {
                {ImmutableList.of("applied-permissions/admin"), true},
                {ImmutableList.of("api:*", "applied-permissions/admin"), true},
                {ImmutableList.of("applied-permissions/nonadmin"), false},
                {ImmutableList.of("applied-permissions/adminn"), false},
                {ImmutableList.of("applied-permissions:admin"), false},
                {ImmutableList.of("applied-permission/admin"), false},
                {ImmutableList.of("applied-permission"), false},
                {ImmutableList.of("admin"), false},
                {ImmutableList.of("applied-permissions/user"), false},
                {ImmutableList.of("jfrt@xyz:admin"), true},
                {ImmutableList.of("api:*", "jfrt@xyz:admin"), true},
                {ImmutableList.of("jfxr@xyz:admin"), false},
                {ImmutableList.of("api:*", "jfxr@xyz:admin"), false},
                {ImmutableList.of("jfrt@abcd:admin"), false},
                {ImmutableList.of("api:*", "jfrt@abcd:admin"), false},
                {ImmutableList.of("jfrt@xyz:user"), false},
        };
    }
}