package org.artifactory.storage.service;

import com.google.common.collect.ImmutableList;
import com.google.common.collect.ImmutableSet;
import com.google.common.collect.Maps;
import org.artifactory.addon.AddonsManager;
import org.artifactory.addon.HaAddon;
import org.artifactory.api.config.CentralConfigService;
import org.artifactory.api.context.ArtifactoryContextThreadBinder;
import org.artifactory.api.repo.RepositoryService;
import org.artifactory.api.storage.BinariesInfo;
import org.artifactory.common.home.ArtifactoryHome;
import org.artifactory.descriptor.config.CentralConfigDescriptor;
import org.artifactory.mbean.MBeanRegistrationService;
import org.artifactory.model.common.RepoPathImpl;
import org.artifactory.model.xstream.fs.FileInfoImpl;
import org.artifactory.projects.ProjectsService;
import org.artifactory.repo.trash.TrashService;
import org.artifactory.sapi.interceptor.context.DeleteContext;
import org.artifactory.schedule.TaskService;
import org.artifactory.spring.InternalArtifactoryContext;
import org.artifactory.storage.StorageSummaryInfo;
import org.artifactory.storage.StorageUsage;
import org.artifactory.storage.binstore.service.InternalBinaryService;
import org.artifactory.storage.db.DbService;
import org.artifactory.storage.db.fs.model.DbMutableFile;
import org.artifactory.storage.fs.repo.CacheUnAvailableException;
import org.artifactory.storage.fs.repo.StoringRepo;
import org.artifactory.storage.fs.service.ConfigsService;
import org.artifactory.storage.fs.service.FileService;
import org.artifactory.storage.jobs.SyncStorageUsageJob;
import org.artifactory.test.ArtifactoryHomeBoundTest;
import org.jfrog.common.JsonUtils;
import org.jfrog.storage.DbType;
import org.jfrog.storage.binstore.ifc.model.BinaryProvidersInfo;
import org.jfrog.storage.binstore.ifc.model.BinaryTreeElement;
import org.joda.time.DateTime;
import org.mockito.Mock;
import org.mockito.MockitoAnnotations;
import org.testng.Assert;
import org.testng.annotations.AfterMethod;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.DataProvider;
import org.testng.annotations.Test;

import java.sql.SQLException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.atomic.AtomicLong;

import static org.mockito.ArgumentMatchers.anyBoolean;
import static org.mockito.ArgumentMatchers.anyLong;
import static org.mockito.ArgumentMatchers.anySet;
import static org.mockito.ArgumentMatchers.anyString;
import static org.mockito.ArgumentMatchers.argThat;
import static org.mockito.ArgumentMatchers.eq;
import static org.mockito.Mockito.doNothing;
import static org.mockito.Mockito.doReturn;
import static org.mockito.Mockito.doThrow;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.spy;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.verifyNoMoreInteractions;
import static org.mockito.Mockito.when;
import static org.testng.Assert.assertEquals;
import static org.testng.Assert.assertNull;
import static org.testng.Assert.assertThrows;

/**
 * @author barh
 */
public class StorageServiceImplTest extends ArtifactoryHomeBoundTest {

    private final static String PROJECTX = "projectx";
    private final static String PROJECTY = "projecty";
    public static final String PROJ_REPO1 = "source-repo";
    public static final String PROJ_REPO2 = "target-repo";

    @Mock
    private CentralConfigService centralConfigService;
    @Mock
    private DbService dbService;
    @Mock
    private InternalBinaryService binaryStore;
    @Mock
    private FileService fileService;
    @Mock
    private TaskService taskService;
    @Mock
    private RepositoryService repositoryService;
    @Mock
    private AddonsManager addonsManager;
    @Mock
    private ProjectsService projectsService;
    @Mock
    private ConfigsService configsService;
    @Mock
    private TrashService trashService;
    @Mock
    InternalArtifactoryContext context;

    @BeforeMethod
    public void init() {
        bindArtifactoryHome();
        MockitoAnnotations.initMocks(this);
        MBeanRegistrationService mbeanService = mock(MBeanRegistrationService.class);
        when(context.beanForType(MBeanRegistrationService.class)).thenReturn(mbeanService);
        ArtifactoryContextThreadBinder.bind(context);
    }

    @AfterMethod
    public void cleanup() {
        ArtifactoryContextThreadBinder.unbind();
        unbindArtifactoryHome();
    }

    @Test
    public void testHideSensitiveInfo() {
        StorageServiceImpl storageHalfMock = spy(StorageServiceImpl.class);
        InternalBinaryService internalBinaryService = mock(InternalBinaryService.class);
        storageHalfMock.setInternalBinaryService(internalBinaryService);
        BinaryProvidersInfo<Map<String, String>> binaryProvidersInfo = createProviderInfo();
        when(internalBinaryService.getBinaryProvidersInfo()).thenReturn(binaryProvidersInfo);
        CentralConfigDescriptor descriptor = mock(CentralConfigDescriptor.class);
        when(centralConfigService.getDescriptor()).thenReturn(descriptor);
        storageHalfMock.setCentralConfigService(centralConfigService);

        BinaryProvidersInfo<Map<String, String>> binaryProviderInfo = storageHalfMock.getBinaryProviderInfo();
        Map<String, String> rootData = binaryProviderInfo.rootTreeElement.getData();
        assertEquals(rootData.size(), 1);
        Assert.assertTrue(rootData.containsKey("other"));
        Assert.assertFalse(rootData.containsKey("identity"));
        Assert.assertFalse(rootData.containsKey("credential"));
        Map<String, String> secondData = binaryProviderInfo.rootTreeElement.getNextBinaryTreeElement().getData();
        assertEquals(secondData.size(), 2);
        Assert.assertTrue(secondData.containsKey("path"));
        Assert.assertTrue(secondData.containsKey("https"));
        Assert.assertFalse(secondData.containsKey("accountKey"));
        Assert.assertFalse(secondData.containsKey("accountName"));
    }

    private BinaryProvidersInfo<Map<String, String>> createProviderInfo() {
        BinaryTreeElement<Map<String, String>> rootElement = new BinaryTreeElement<>();
        HashMap<String, String> data = Maps.newHashMap();
        data.put("credential", "c");
        data.put("identity", "d");
        data.put("other", "d");
        rootElement.setData(data);

        BinaryTreeElement<Map<String, String>> secondElement = new BinaryTreeElement<>();
        HashMap<String, String> secondData = Maps.newHashMap();
        secondData.put("accountKey", "k");
        secondData.put("accountName", "n");
        secondData.put("path", "b");
        secondData.put("https", "e");
        secondElement.setData(secondData);

        rootElement.setNextBinaryTreeElement(secondElement);
        return new BinaryProvidersInfo<>("cache", rootElement);
    }

    @Test
    void testGetCachedStorageSizeResolveFromDB() throws SQLException {
        //prepare
        StorageServiceImpl storageHalfMock = spy(StorageServiceImpl.class);
        doThrow(CacheUnAvailableException.class).when(storageHalfMock).getStorageSummaryInfoFromCache();
        InternalBinaryService internalBinaryService = mock(InternalBinaryService.class);
        when(internalBinaryService.getBinariesInfo()).thenReturn(mock(BinariesInfo.class));
        storageHalfMock.setInternalBinaryService(internalBinaryService);

        //act
        when(storageHalfMock.getCachedStorageSize()).thenCallRealMethod();

        //verify
        verify(internalBinaryService, times(1)).getBinariesInfo();
    }

    @Test
    void testGetCachedStorageSizeResolveFromCache() throws SQLException {
        //prepare
        StorageServiceImpl storageHalfMock = spy(StorageServiceImpl.class);
        StorageSummaryInfo storageSummaryInfo = mock(StorageSummaryInfo.class);
        when(storageSummaryInfo.getBinariesInfo()).thenReturn(mock(BinariesInfo.class));
        doReturn(storageSummaryInfo).when(storageHalfMock).getStorageSummaryInfoFromCache();
        InternalBinaryService internalBinaryService = mock(InternalBinaryService.class);
        storageHalfMock.setInternalBinaryService(internalBinaryService);

        //act
        when(storageHalfMock.getCachedStorageSize()).thenCallRealMethod();

        //verify
        verify(internalBinaryService, times(0)).getBinariesInfo();
    }

    @Test
    void testGetProjectStorageSum() {
        StorageServiceImpl storageService = initStorageService();
        assertEquals(1000L, storageService.getProjectStorageUsage(PROJECTX));
        assertEquals(0L, storageService.getProjectStorageUsage(ProjectsService.DEFAULT_PROJECT));
    }

    @Test
    void testSyncStorageUsageDiffsIfNeeded() {
        // Setup
        StorageServiceImpl storageService = initStorageService();
        when(repositoryService.getAllRepoKeys()).thenReturn(ImmutableSet.of(PROJ_REPO1));
        when(projectsService.getAllProjects()).thenReturn(List.of(PROJECTX));
        when(projectsService.getProjectKeyOfRepository(PROJ_REPO1)).thenReturn(PROJECTX);
        doThrow(IllegalArgumentException.class).when(storageService).saveAndReloadStorageUsageDiffsWithRetries();
        HaAddon haAddon = mock(HaAddon.class);
        when(haAddon.getAllArtifactoryServers()).thenReturn(new ArrayList<>());
        when(addonsManager.addonByType(HaAddon.class)).thenReturn(haAddon);

        // Set the project limit for 2000 bytes
        when(projectsService.getProjectStorageQuotaBytes(PROJECTX)).thenReturn(2000L);
        // Call syncStorageUsageDiffsIfNeeded and expect nothing to happened - tests the flush threshold
        storageService.syncStorageUsageDiffsIfNeeded();

        //Deploy a small file into a project related repository
        storageService.afterCreate(prepareDbMutableFileMock(1), null, null);
        // Call syncStorageUsageDiffsIfNeeded and expect nothing to happened
        storageService.syncStorageUsageDiffsIfNeeded();

        // Deploy a bigger file into the project related repository
        storageService.afterCreate(prepareDbMutableFileMock(500), null, null);
        try {
            // Call syncStorageUsageDiffsIfNeeded and expect saveAndReloadStorageUsageDiffsWithRetries will get called
            storageService.syncStorageUsageDiffsIfNeeded();
            Assert.fail("saveAndReloadStorageUsageDiffsWithRetries was never called");
        } catch (IllegalArgumentException ignored) {}

        when(trashService.isTrashcanEnabled()).thenReturn(false);
        // Delete the file from before so the total diff will be only 1 byte
        DeleteContext deleteContext = new DeleteContext(new RepoPathImpl(PROJ_REPO1, "test.file"));
        storageService.afterDelete(prepareDbMutableFileMock(500), null, deleteContext);
        // Call syncStorageUsageDiffsIfNeeded for 15 times in total (including the previous retries) and expect
        // saveAndReloadStorageUsageDiffsWithRetries will get called at the last try
        for (int i = 0; i < 11; i++) {
            storageService.syncStorageUsageDiffsIfNeeded();
        }
        try {
            storageService.syncStorageUsageDiffsIfNeeded();
            Assert.fail("saveAndReloadStorageUsageDiffsWithRetries was never called");
        } catch (IllegalArgumentException ignored) {}
    }

    @Test
    void testUpdateStorageSummaryWithStorageUsage() {
        StorageServiceImpl storageService = initStorageService();
        StorageUsage storageUsage = new StorageUsage();
        storageUsage.getProjectsStorageSum().put("project", 2000L);
        storageUsage.getRepoStorageSum().put("proj-repo", 1000L);
        storageUsage.getRepoStorageSum().put("proj-remote-cache", 1000L);

        Map<String, String> repoToProjMap = new HashMap<>();
        repoToProjMap.put("proj-repo", "project");
        repoToProjMap.put("proj-remote", "project");
        when(projectsService.getRepoKeyToProjectMap(anySet(), anyBoolean())).thenReturn(repoToProjMap);

        when(repositoryService.getAllRepoKeys())
                .thenReturn(ImmutableSet.of("proj-repo", "proj-remote-cache", "proj-remote"));

        storageService.updateStorageSummaryWithStorageUsage(storageUsage);

        StorageSummaryInfo cache = storageService.getStorageSummaryInfoFromCache();
        assertEquals(cache.getProjectsTotalSizes().getOrDefault("project", 0L).longValue(), 2000L);
        assertEquals(cache.getRepoStorageSummaries().get("proj-repo").getUsedSpace(), 1000L);
        assertEquals(cache.getRepoStorageSummaries().get("proj-remote-cache").getProjectKey(), "project");
    }

    @Test(description = "replicates a scenario where all repos are present, but one switched a project")
    void testMergeStorageDiffWithStorageUsageWithAllReposPresent() {
        String otherProjRepo = "other-proj-repo";
        StorageServiceImpl storageService = initStorageService();
        when(projectsService.getProjectKeyOfRepository(PROJ_REPO1)).thenReturn(PROJECTX);
        when(projectsService.getProjectKeyOfRepository(otherProjRepo)).thenReturn(PROJECTX);
        Map<String, String> repoToProjMap = new HashMap<>();
        repoToProjMap.put(PROJ_REPO1, PROJECTX);
        repoToProjMap.put(otherProjRepo, PROJECTX);
        when(projectsService.getRepoKeyToProjectMap(anySet(), anyBoolean())).thenReturn(repoToProjMap);
        when(repositoryService.getAllRepoKeys()).thenReturn(ImmutableSet.of(otherProjRepo, PROJ_REPO1));

        StorageUsage oldStorageUsage = new StorageUsage();
        oldStorageUsage.setRevision(1);
        oldStorageUsage.getProjectsStorageSum().put(PROJECTX, 1000L);
        oldStorageUsage.getRepoStorageSum().put(otherProjRepo, 1000L);

        // Deploy some file so we gonna have some changes to merge
        storageService.afterCreate(prepareDbMutableFileMock(500), null, null);

        // Call mergeStorageDiffWithStorageUsage
        StorageUsage mergedStorageUsage = storageService.mergeStorageDiffWithStorageUsage(oldStorageUsage);

        // Verify the merged StorageUsage contains both local changes and oldStorageUsage
        assertEquals(mergedStorageUsage.getRepoStorageSum().get(PROJ_REPO1).longValue(), 500);
        assertEquals(mergedStorageUsage.getRepoStorageSum().get(otherProjRepo).longValue(), 1000);
        assertEquals(mergedStorageUsage.getProjectsStorageSum().get(PROJECTX).longValue(), 1500);
        assertEquals(mergedStorageUsage.getRevision(), 2);
    }

    @Test(description = "replicates a scenario where one repo was removed, but a file was added to a previous project" +
            " size")
    void testMergeStorageDiffWithStorageUsageWithOneRepoRemoved() {
        // repo below will be deleted
        String otherProjRepo = "other-proj-repo";
        StorageServiceImpl storageService = initStorageService();
        // repository service indicates that only one repo exists now
        when(repositoryService.getAllRepoKeys()).thenReturn(ImmutableSet.of(PROJ_REPO1));
        when(projectsService.getProjectKeyOfRepository(PROJ_REPO1)).thenReturn(PROJECTX);
        // below simulates response from Access
        Map<String, String> repoToProjMap = new HashMap<>();
        repoToProjMap.put(PROJ_REPO1, PROJECTX);
        when(projectsService.getRepoKeyToProjectMap(anySet(), anyBoolean())).thenReturn(repoToProjMap);

        StorageUsage oldStorageUsage = new StorageUsage();
        oldStorageUsage.setRevision(1);
        oldStorageUsage.getProjectsStorageSum().put(PROJECTX, 1000L);
        oldStorageUsage.getRepoStorageSum().put(otherProjRepo, 1000L);

        // Deploy some file so we gonna have some changes to merge
        storageService.afterCreate(prepareDbMutableFileMock(500), null, null);

        // Call mergeStorageDiffWithStorageUsage
        StorageUsage mergedStorageUsage = storageService.mergeStorageDiffWithStorageUsage(oldStorageUsage);

        // Verify the merged StorageUsage contains both local changes and oldStorageUsage
        assertEquals(mergedStorageUsage.getRepoStorageSum().get(PROJ_REPO1).longValue(), 500);
        assertNull(mergedStorageUsage.getRepoStorageSum().get(otherProjRepo));
        assertEquals(mergedStorageUsage.getProjectsStorageSum().get(PROJECTX).longValue(), 500);
        assertEquals(mergedStorageUsage.getRevision(), 2);
    }

    @Test
    void testMergeStorageDiffWithStorageUsageWithRepoReassignedOnlyAndWithoutDiffs() {
        StorageUsage oldStorageUsage = new StorageUsage();
        oldStorageUsage.setRevision(1);
        oldStorageUsage.getProjectsStorageSum().put(PROJECTX, 1000L);
        oldStorageUsage.getProjectsStorageSum().put(PROJECTY, 1000L);
        oldStorageUsage.getRepoStorageSum().put(PROJ_REPO1, 1000L);
        oldStorageUsage.getRepoStorageSum().put(PROJ_REPO2, 1000L);

        StorageServiceImpl storageService = initStorageService();
        when(projectsService.getAllProjects()).thenReturn(ImmutableList.of(PROJECTX, PROJECTY));
        when(repositoryService.getAllRepoKeys()).thenReturn(ImmutableSet.of(PROJ_REPO2, PROJ_REPO1));
        Map<String, String> repoToProjMap = new HashMap<>();
        repoToProjMap.put(PROJ_REPO1, PROJECTX);
        repoToProjMap.put(PROJ_REPO2, PROJECTX);
        when(projectsService.getRepoKeyToProjectMap(anySet(), anyBoolean())).thenReturn(repoToProjMap);

        // Call mergeStorageDiffWithStorageUsage
        StorageUsage mergedStorageUsage = storageService.mergeStorageDiffWithStorageUsage(oldStorageUsage);

        // Verify the merged StorageUsage contains both local changes and oldStorageUsage
        assertEquals(mergedStorageUsage.getProjectsStorageSum().get(PROJECTX).longValue(), 2000);
        assertEquals(mergedStorageUsage.getProjectsStorageSum().get(PROJECTY).longValue(), 0);
        assertEquals(mergedStorageUsage.getRepoStorageSum().get(PROJ_REPO1).longValue(), 1000);
        assertEquals(mergedStorageUsage.getRepoStorageSum().get(PROJ_REPO2).longValue(), 1000);
        assertEquals(mergedStorageUsage.getRevision(), 2);
    }

    @Test
    void testSyncStorageUsageWithStorageSummary() {
        StorageServiceImpl storageService = initStorageService();
        when(repositoryService.getAllRepoKeys()).thenReturn(ImmutableSet.of("example-repo-local", "prjx-generic",
                "artifactory-build-info", "auto-trashcan", "jfrog-support-bundle"));
        when(configsService.getConfig(ArtifactoryHome.STORAGE_USAGE_FILE)).thenReturn("{\"revision\":1603365765723," +
                "\"repoStorageSum\":{\"example-repo-local\":5703478,\"prjx-generic\":43741906," +
                "\"artifactory-build-info\":0,\"auto-trashcan\":0,\"jfrog-support-bundle\":0}," +
                "\"projectsStorageSum\":{\"default\":5703478,\"projectx\":43741906}}");
        // Sync
        storageService.syncStorageUsageWithStorageSummary();

        // Verify the local cache has been updated correctly
        StorageSummaryInfo cache = storageService.getStorageSummaryInfoFromCache();
        assertEquals(43741906, cache.getProjectsTotalSizes().get(PROJECTX).longValue());
        assertEquals(43741906, cache.getRepoStorageSummaries().get("prjx-generic").getUsedSpace());
        assertEquals(5703478, cache.getProjectsTotalSizes().get(ProjectsService.DEFAULT_PROJECT).longValue());
        assertEquals(5703478, cache.getRepoStorageSummaries().get("example-repo-local").getUsedSpace());
        assertEquals(0, cache.getRepoStorageSummaries().get("auto-trashcan").getUsedSpace());
    }

    @Test
    void testMergeDiffChanges() {
        StorageServiceImpl storageService = initStorageService();
        Map<String, AtomicLong> realDiffMap = new ConcurrentHashMap<>();
        Map<String, AtomicLong> backupDiffMap = new HashMap<>();
        // Create some data to merge
        realDiffMap.put("repo", new AtomicLong(10000L));
        backupDiffMap.put("repo", new AtomicLong(99L));
        backupDiffMap.put("repo2", new AtomicLong(100L));
        // Act
        storageService.mergeDiffChanges(backupDiffMap, realDiffMap);
        // Validate the merged data contains both source and target maps
        assertEquals(realDiffMap.get("repo").get(), 10099L);
        assertEquals(realDiffMap.get("repo2").get(), 100L);
    }

    @Test
    void testGetAllProjectsNotifiedOnStorageQuota() {
        StorageServiceImpl storageService = initStorageService();
        when(configsService.getConfig(ArtifactoryHome.STORAGE_USAGE_FILE)).thenReturn(null);

        // storage usage info must not be empty
        assertThrows(IllegalStateException.class, storageService::getAllProjectsNotifiedOnStorageQuota);

        Map<String, Long> notifiedProjects = Map.of("p1", 0L, "p2", 1L);
        StorageUsage storageUsage = new StorageUsage();
        storageUsage.setProjectsLastNotifiedOnStorageQuota(notifiedProjects);
        String storageUsageString = JsonUtils.getInstance().writeValueAsString(storageUsage);
        when(configsService.getConfig(ArtifactoryHome.STORAGE_USAGE_FILE)).thenReturn(storageUsageString);

        Map<String, Long> notifiedProjectsResult = storageService.getAllProjectsNotifiedOnStorageQuota();
        assertEquals(notifiedProjectsResult.size(), 2);
    }

    @Test
    void testUpdateProjectStorageQuotaNotificationTimestamp() {
        Set<String> projects = Set.of("p1");
        StorageServiceImpl storageService = initStorageService();
        StorageUsage storageUsage = new StorageUsage();
        when(configsService.getConfig(ArtifactoryHome.STORAGE_USAGE_FILE))
                .thenReturn(JsonUtils.getInstance().writeValueAsString(storageUsage));
        when(configsService
                .updateConfigIfLastModificationMatch(eq(ArtifactoryHome.STORAGE_USAGE_FILE), anyString(), anyLong(),
                        anyLong()))
                .thenReturn(false)
                .thenReturn(true);

        storageService.updateProjectStorageQuotaNotificationTimestamp(projects);

        verify(configsService, times(2)).updateConfigIfLastModificationMatch(eq(ArtifactoryHome.STORAGE_USAGE_FILE),
                argThat(storageUsageString -> {
                    StorageUsage storageUsageArgument =
                            JsonUtils.getInstance().readValue(storageUsageString, StorageUsage.class);

                    long timeDelta =
                            DateTime.now().minus(storageUsageArgument.getProjectsLastNotifiedOnStorageQuota().get("p1"))
                                    .getMillis();
                    boolean projectLastNotifiedIsCloseToNow = timeDelta < 5000; // within 5 seconds range
                    return storageUsageArgument.getProjectsLastNotifiedOnStorageQuota().size() == 1 &&
                            projectLastNotifiedIsCloseToNow;
                }), eq(1L), eq(2L));
    }

    @Test(dataProvider = "provideEntLicense")
    void testActivateStorageUsageSyncJobIfNeeded(boolean isEntLicense) {
        StorageServiceImpl storageService = initStorageService();
        when(addonsManager.isEnterpriseLicensed()).thenReturn(isEntLicense);

        storageService.onLicenseLoaded();

        if (isEntLicense) {
            verify(taskService).startTask(
                    argThat((task) -> task.getType().equals(SyncStorageUsageJob.class)), eq(false));
        }
        verifyNoMoreInteractions(taskService);
    }

    private StorageServiceImpl initStorageService() {
        StorageServiceImpl storageService = spy(new StorageServiceImpl());
        storageService.setAddonsManager(addonsManager);
        storageService.setBinaryStore(binaryStore);
        storageService.setCentralConfigService(centralConfigService);
        storageService.setDbService(dbService);
        storageService.setFileService(fileService);
        storageService.setTaskService(taskService);
        storageService.setRepositoryService(repositoryService);
        storageService.setProjectsService(projectsService);
        storageService.setConfigsService(configsService);
        storageService.setTrashService(trashService);
        when(dbService.getDatabaseType()).thenReturn(DbType.MYSQL);
        CentralConfigDescriptor centralConfigDescriptor = mock(CentralConfigDescriptor.class);
        when(centralConfigService.getDescriptor()).thenReturn(centralConfigDescriptor);
        doNothing().when(storageService).rescheduleGC(centralConfigDescriptor, null);
        doReturn(true).when(storageService).isProjectsStorageQuotaEnabled();
        storageService.init();
        populateStorageSummaryCache();
        storageService.updateStorageSummaryCache(populateStorageSummaryCache());
        return storageService;
    }

    private StorageSummaryInfo populateStorageSummaryCache() {
        StorageSummaryInfo summaryInfo =
                new StorageSummaryInfo(new HashMap<>(), null, new HashMap<>(), new HashMap<>());
        summaryInfo.getProjectsTotalSizes().putIfAbsent(PROJECTX, 1000L);
        return summaryInfo;
    }

    public static DbMutableFile prepareDbMutableFileMock(long size) {
        RepoPathImpl repoPath = new RepoPathImpl(PROJ_REPO1, "test.file");
        FileInfoImpl fileInfo = new FileInfoImpl(repoPath);
        fileInfo.setSize(size);
        return new DbMutableFile(mock(StoringRepo.class), 52437842367L, fileInfo);
    }

    @DataProvider
    public Object[][] provideEntLicense() {
        return new Object[][] {
                {true},
                {false}
        };
    }
}