/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2018 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.ui.rest.service.artifacts.browse.treebrowser.tree;

import com.fasterxml.jackson.annotation.JsonInclude;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.ObjectWriter;
import com.fasterxml.jackson.databind.introspect.JacksonAnnotationIntrospector;
import org.artifactory.api.rest.common.model.continues.ContinueResult;
import org.artifactory.common.ConstantValues;
import org.artifactory.projects.ProjectsService;
import org.artifactory.rest.common.model.RestModel;
import org.artifactory.rest.common.service.ArtifactoryRestRequest;
import org.artifactory.rest.common.service.RestResponse;
import org.artifactory.rest.common.service.RestService;
import org.artifactory.ui.rest.model.artifacts.browse.treebrowser.nodes.RestTreeNode;
import org.artifactory.ui.rest.model.artifacts.browse.treebrowser.nodes.repo.RepositoryNode;
import org.jfrog.access.client.project.ProjectResourceEntity;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.config.BeanDefinition;
import org.springframework.context.annotation.Scope;
import org.springframework.stereotype.Component;

import javax.ws.rs.HttpMethod;
import javax.ws.rs.core.StreamingOutput;
import java.io.BufferedWriter;
import java.io.IOException;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.util.HashMap;
import java.util.Map;

/**
 * @author Chen Keinan
 */
@Component
@Scope(BeanDefinition.SCOPE_PROTOTYPE)
public class BrowseTreeNodesService implements RestService<RestTreeNode> {

    private ProjectsService projectsService;

    @Autowired
    public void setProjectsService(ProjectsService projectsService) {
        this.projectsService = projectsService;
    }

    @Override
    public void execute(ArtifactoryRestRequest<RestTreeNode> request, RestResponse response) {
        boolean isCompact = Boolean.parseBoolean(request.getQueryParamByKey("compacted")) &&
                ConstantValues.treebrowserFolderCompact.getBoolean();
        ContinueResult<? extends RestModel> items = fetchItemTypeData(request, isCompact);
        if (isArchiveRequest(request)) {
            StreamingOutput streamingOutput = os -> write(items, os);
            response.iModel(streamingOutput);
            return;
        }
        addProjectRelatedData(items);
        response.iModel(items);
    }

    ContinueResult<? extends RestModel> fetchItemTypeData(ArtifactoryRestRequest<RestTreeNode> request,
            boolean isCompact) {
        return request.getImodel().fetchItemTypeData(isCompact);
    }

    private void addProjectRelatedData(ContinueResult<? extends RestModel> items) {
        Map<String, ProjectResourceEntity> resources = getProjectResources();
        items.getData().forEach(model -> setSharingProperties(resources, model));
    }

    private Map<String, ProjectResourceEntity> getProjectResources() {
        Map<String, Map<String, ProjectResourceEntity>> projectsResources =
                projectsService.getAllProjectsRepositoriesMap(false);
        Map<String, ProjectResourceEntity> entities = new HashMap<>();
        projectsResources.values().forEach(entities::putAll);
        return entities;
    }

    private void setSharingProperties(Map<String, ProjectResourceEntity> entities, RestModel model) {
        if (model instanceof RepositoryNode) {
            RepositoryNode node = (RepositoryNode) model;
            String key = node.getRepoKey();
            ProjectResourceEntity entity = entities.get(key);
            if (entity != null) {
                node.setSharedReadOnly(entity.isSharedReadOnly());
                node.setTarget(entity.getResourceStatus().isTarget());
            }
        }
    }

    void write(ContinueResult<? extends RestModel> items, OutputStream os) throws IOException {
        Writer writer = new BufferedWriter(new OutputStreamWriter(os));
        ObjectWriter jsonObjectWriter = new ObjectMapper()
                .setSerializationInclusion(JsonInclude.Include.NON_NULL)
                .setAnnotationIntrospector(new JacksonAnnotationIntrospector())
                .writer();
        jsonObjectWriter.writeValue(writer, items);
    }

    boolean isArchiveRequest(ArtifactoryRestRequest<RestTreeNode> request) {
        return request.getRequest().getMethod().equalsIgnoreCase(HttpMethod.POST) &&
                request.getImodel().isArchiveExpendRequest();
    }
}