package org.artifactory.ui.rest.service.artifacts.browse.treebrowser.tree;

import com.fasterxml.jackson.databind.JsonNode;
import org.apache.commons.compress.archivers.zip.ZipArchiveInputStream;
import org.artifactory.api.jackson.JacksonReader;
import org.artifactory.api.rest.common.model.continues.ContinueResult;
import org.artifactory.descriptor.repo.LocalRepoDescriptor;
import org.artifactory.descriptor.repo.RepoType;
import org.artifactory.fs.ItemInfo;
import org.artifactory.model.common.RepoPathImpl;
import org.artifactory.model.xstream.fs.FileInfoImpl;
import org.artifactory.projects.ProjectsService;
import org.artifactory.rest.common.model.RestModel;
import org.artifactory.rest.common.service.ArtifactoryRestRequest;
import org.artifactory.rest.common.service.ArtifactoryRestResponse;
import org.artifactory.ui.rest.model.artifacts.browse.treebrowser.nodes.JunctionNode;
import org.artifactory.ui.rest.model.artifacts.browse.treebrowser.nodes.RestTreeNode;
import org.artifactory.ui.rest.model.artifacts.browse.treebrowser.nodes.archive.ArchiveEntriesTree;
import org.artifactory.ui.rest.model.artifacts.browse.treebrowser.nodes.repo.RepositoryNode;
import org.artifactory.ui.utils.MockArtifactoryArchiveEntriesTree;
import org.jfrog.access.client.project.ProjectResourceEntity;
import org.jfrog.access.client.project.ProjectResourceStatus;
import org.jfrog.common.ResourceUtils;
import org.mockito.Mockito;
import org.testng.annotations.Test;

import javax.ws.rs.HttpMethod;
import javax.ws.rs.core.Request;
import javax.ws.rs.core.StreamingOutput;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.List;
import java.util.Map;

import static org.apache.http.HttpStatus.SC_OK;
import static org.artifactory.repo.RepoDetailsType.FEDERATED_REPO;
import static org.artifactory.repo.RepoDetailsType.LOCAL_REPO;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.when;
import static org.testng.Assert.assertEquals;
import static org.testng.Assert.assertFalse;
import static org.testng.Assert.assertTrue;

/**
 * @author Alexei Vainshtein
 */
public class BrowseTreeNodesServiceTest extends MockArtifactoryArchiveEntriesTree {

    private static final String ZIP_FILE_NAME = "test.zip";
    private static final String TEST_ZIP = "/browse/treebrowser/" + ZIP_FILE_NAME;
    private static final String SHARED_READ_ONLY_REPO = "sharedReadOnlyRepo";
    private static final String NOT_SHARED_READ_ONLY_REPO = "notSharedReadOnlyRepo";

    @Test
    public void testExecuteAsArchive() throws IOException {
        File archive = ResourceUtils.getResourceAsFile(TEST_ZIP);
        prepareMock(ZIP_FILE_NAME, LOCAL_REPO);
        InputStream ARCHIVE_STREAM = new FileInputStream(archive);
        when(repoService.archiveInputStream(any())).thenReturn(new ZipArchiveInputStream(ARCHIVE_STREAM));
        ArtifactoryRestResponse response = new ArtifactoryRestResponse();
        BrowseTreeNodesService browseTreeNodesService = new BrowseTreeNodesService();
        assertTrue(browseTreeNodesService.isArchiveRequest(artifactoryRestRequest));
        browseTreeNodesService.execute(artifactoryRestRequest, response);
        assertTrue(response.getIModel() instanceof StreamingOutput);
        assertEquals(response.buildResponse().getStatus(), SC_OK);
    }

    @Test
    public void testExecuteArchiveRequestForFederatedRepo() throws IOException {
        File archive = ResourceUtils.getResourceAsFile(TEST_ZIP);
        prepareMock(ZIP_FILE_NAME, FEDERATED_REPO);
        InputStream ARCHIVE_STREAM = new FileInputStream(archive);
        when(repoService.archiveInputStream(any())).thenReturn(new ZipArchiveInputStream(ARCHIVE_STREAM));
        ArtifactoryRestResponse response = new ArtifactoryRestResponse();
        BrowseTreeNodesService browseTreeNodesService = new BrowseTreeNodesService();
        assertTrue(browseTreeNodesService.isArchiveRequest(artifactoryRestRequest));
        browseTreeNodesService.execute(artifactoryRestRequest, response);
        assertTrue(response.getIModel() instanceof StreamingOutput);
        assertEquals(response.buildResponse().getStatus(), SC_OK);
    }

    @Test
    public void testExecuteNonArchiveRequest() {
        LocalRepoDescriptor repoDescriptor = new LocalRepoDescriptor();
        repoDescriptor.setType(RepoType.Generic);
        prepareMock("test.txt", LOCAL_REPO);
        when(context.getRepositoryService()).thenReturn(repoService);
        when(repoService.repoDescriptorByKey(any())).thenReturn(repoDescriptor);
        ArtifactoryRestResponse response = new ArtifactoryRestResponse();
        BrowseTreeNodesService browseTreeNodesService = new BrowseTreeNodesService();
        ProjectsService projectsService = Mockito.mock(ProjectsService.class);
        browseTreeNodesService.setProjectsService(projectsService);
        assertFalse(browseTreeNodesService.isArchiveRequest(artifactoryRestRequest));
        browseTreeNodesService.execute(artifactoryRestRequest, response);
        assertTrue(response.getIModel() instanceof ContinueResult);
        assertEquals(response.buildResponse().getStatus(), SC_OK);
    }

    @Test
    public void testSharedReadonlyFlag() {
        prepareMock("test.txt", LOCAL_REPO);
        ArtifactoryRestResponse response = new ArtifactoryRestResponse();
        BrowseTreeNodesService browseTreeNodesService = new BrowseTreeNodesServiceForTest();
        ProjectsService projectsService = prepareProjectsServiceMock();
        browseTreeNodesService.setProjectsService(projectsService);

        browseTreeNodesService.execute(artifactoryRestRequest, response);

        List<RepositoryNode> list = ((ContinueResult) response.getIModel()).getData();
        RepositoryNode shareReadOnlyNode = list.stream()
                .filter(o -> o.getRepoKey().equals(SHARED_READ_ONLY_REPO)).findFirst().get();
        RepositoryNode notSharedReadOnlyNode = list.stream()
                .filter(o -> o.getRepoKey().equals(NOT_SHARED_READ_ONLY_REPO)).findFirst().get();
        assertTrue(shareReadOnlyNode.isSharedReadOnly());
        assertTrue(shareReadOnlyNode.isTarget());
        assertFalse(notSharedReadOnlyNode.isSharedReadOnly());
        assertFalse(notSharedReadOnlyNode.isTarget());
    }

    private ProjectsService prepareProjectsServiceMock() {
        ProjectsService projectsService = Mockito.mock(ProjectsService.class);
        ProjectResourceEntity sharedReadOnlyEntity = ProjectResourceEntity.builder()
                .sharedReadOnly(true)
                .resourceStatus(ProjectResourceStatus.SHARED_TARGET)
                .build();
        ProjectResourceEntity notSharedReadOnlyEntity = ProjectResourceEntity.builder()
                .sharedReadOnly(false)
                .resourceStatus(ProjectResourceStatus.SHARED)
                .build();
        when(projectsService.getAllProjectsRepositoriesMap(false))
                .thenReturn(Map.of("a", Map.of(
                        SHARED_READ_ONLY_REPO, sharedReadOnlyEntity,
                        NOT_SHARED_READ_ONLY_REPO, notSharedReadOnlyEntity
                )));
        return projectsService;
    }

    @Test
    public void testWrite() throws IOException {
        File archive = ResourceUtils.getResourceAsFile(TEST_ZIP);
        OutputStream output = new OutputStream() {
            private final StringBuilder string = new StringBuilder();

            @Override
            public void write(int x) {
                this.string.append((char) x);
            }

            public String toString() {
                return this.string.toString();
            }
        };
        InputStream ARCHIVE_STREAM = new FileInputStream(archive);
        when(repoService.archiveInputStream(any())).thenReturn(new ZipArchiveInputStream(ARCHIVE_STREAM));
        BrowseTreeNodesService browseTreeNodesService = new BrowseTreeNodesService();
        ContinueResult<? extends RestModel> continueResult =
                new ArchiveEntriesTree().buildChildren("test", ZIP_FILE_NAME);
        browseTreeNodesService.write(continueResult, output);
        String streamResult = output.toString();
        JsonNode jsonNode = JacksonReader.bytesAsTree(streamResult.getBytes());
        assertEquals(jsonNode.size(), 1);
        JsonNode data = jsonNode.get("data");
        assertEquals(data.size(), 4);
        assertEquals(data.get(0).get("children").size(), 2);
        assertTrue(data.get(0).get("folder").asBoolean());
    }


    private void prepareMock(String fileName, String repoType) {
        ItemInfo fileInfo = new FileInfoImpl(new RepoPathImpl("test", fileName));
        JunctionNode junctionNode = new JunctionNode();
        junctionNode.setRepoKey("test");
        junctionNode.setRepoType(repoType);
        junctionNode.setPath(fileName);
        when(repoService.getItemInfo(any())).thenReturn(fileInfo);
        when(artifactoryRestRequest.getImodel()).thenReturn(junctionNode);
        Request request = Mockito.mock(Request.class);
        when(artifactoryRestRequest.getRequest()).thenReturn(request);
        when(request.getMethod()).thenReturn(HttpMethod.POST);
    }

    private static class BrowseTreeNodesServiceForTest extends BrowseTreeNodesService {
        @Override
        ContinueResult<? extends RestModel> fetchItemTypeData(ArtifactoryRestRequest<RestTreeNode> request,
                boolean isCompact) {
            RepositoryNode sharedReadOnlyRepoModel =
                    new RepositoryNode(SHARED_READ_ONLY_REPO, RepoType.Generic, "repoType");
            RepositoryNode notSharedReadOnlyRepoModel =
                    new RepositoryNode(NOT_SHARED_READ_ONLY_REPO, RepoType.Generic, "repoType");
            return new ContinueResult<RestModel>(0,
                    List.of(sharedReadOnlyRepoModel, notSharedReadOnlyRepoModel));
        }
    }

}